#include <winsock2.h>
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "NQDSProto.h"
#include "NQDSSrv.h"
#include "NQDSThread.h"

static DWORD WINAPI BroadcastThreadProc(  LPVOID lpParameter );

CBroadcastThread::CBroadcastThread()
{
	m_hThread = NULL;
	m_nThreadID = NULL;
	m_bTerminateSignal = FALSE;
	m_bTerminated = TRUE;

	m_hSocket = INVALID_SOCKET;
	m_hIOEvent = NULL;

	m_nPacketsSum = 0;
	m_nPacketsLost = 0;
	m_nLastSeqNum = 0;

	for ( int i=0; i<BROADCAST_QUEUE_SIZE; i++ )
	{
		m_PacketsQueue[i].Type = 'Q';
		m_PacketsQueue[i].Bid[5] = '.';
		m_PacketsQueue[i].Ask[5] = '.';
	}
}

CBroadcastThread::~CBroadcastThread()
{
	if ( !m_bTerminated )
		SafeTerminateAndCleanUp();
}

BOOL CBroadcastThread::CreateThread( in_addr addrSelf, WORD nPort )
{
	// Init synchronization objects
	for ( int i=0; i<BROADCAST_QUEUE_SIZE; i++ )
		InitializeCriticalSection( &m_QueueLocks[i] );
	InitializeCriticalSection( &m_hCriticalSection );
	m_hSemaphore = CreateSemaphore( NULL, 0, BROADCAST_QUEUE_SIZE, NULL );
	m_nQueueHead = m_nQueueTail = 0;
	// Init socket first
	// Allocate Socket
	m_hSocket = WSASocket(AF_INET, SOCK_DGRAM, IPPROTO_UDP, 
       	0, 0, WSA_FLAG_OVERLAPPED);
   	if ( INVALID_SOCKET == m_hSocket )
   	{
		TRACE( "Error initializing socket!\n" );
   		return FALSE;
   	}
	// Bind address
	SOCKADDR_IN sockAddr;
	memset((char *)&sockAddr, 0, sizeof(sockAddr));
   	sockAddr.sin_family = AF_INET;
   	sockAddr.sin_port = htons((WORD)(nPort));
	sockAddr.sin_addr.s_addr = addrSelf.S_un.S_addr;
   	if( bind(m_hSocket, (SOCKADDR *)&sockAddr, sizeof(sockAddr)) )
   	{
   		TRACE( "Error binding address!\n" );
       	return FALSE;
    }
	// Save port
	m_nPort = nPort;
	// Create IO Synchronization event ( Manual reset, initially unsignaled, nameless object )
	m_hIOEvent = WSACreateEvent();	//CreateEvent( NULL, TRUE, FALSE, NULL );
    if (NULL == m_hIOEvent)
    {
   		TRACE( "Error creating event for socket!\n" );
       	return FALSE;
    }
	// Init Thread
	m_bTerminated = FALSE;
	m_hThread = ::CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)&BroadcastThreadProc, 
		this, 0, &m_nThreadID );
	if ( NULL == m_hThread )
	{
		m_bTerminated = TRUE;
		TRACE( "Error initializing thread!\n" );
		return FALSE;
	}
	return TRUE;
}

void CBroadcastThread::InitInstance()
{
	// TODO: Add your implementation here...
	char dataBuf[UDP_MAX+1];
	WSAOVERLAPPED signal;
	memset( &signal, 0, sizeof( WSAOVERLAPPED ) );
	signal.hEvent = m_hIOEvent;
	WSABUF bufInfo;
	bufInfo.len = UDP_MAX+1;
	bufInfo.buf = dataBuf;

	DWORD nSent;
	DWORD nWaitResult;

	// Prepare broadcast address
	SOCKADDR_IN broadcastAddr;
	memset((char *)&broadcastAddr, 0, sizeof(broadcastAddr));
   	broadcastAddr.sin_family = AF_INET;
   	broadcastAddr.sin_port = htons(m_nPort);
  	broadcastAddr.sin_addr.s_addr = INADDR_BROADCAST;
	// added by Louis ***********************
	
	BOOL bBroadcast=TRUE;
	setsockopt(m_hSocket,SOL_SOCKET,SO_BROADCAST,(LPSTR)&bBroadcast,sizeof(bBroadcast));
	//*****************************************************************
	while ( !m_bTerminateSignal )
	{
		while( WAIT_OBJECT_0 != (nWaitResult = WaitForSingleObject( m_hSemaphore, NQDS_TIMEOUT ) ) )
		{
			if ( m_bTerminateSignal )
				return;
		}
		EnterCriticalSection( &m_QueueLocks[m_nQueueHead] );
		bufInfo.buf = (char *)(&m_PacketsQueue[m_nQueueHead]);
		bufInfo.len = sizeof( QUOT_BROADCAST );
		//TRACE( "%.10s\n", m_PacketsQueue[m_nQueueHead].Seq );
		int nRet = WSASendTo( m_hSocket, &bufInfo, 1, &nSent, NULL, (sockaddr *)&broadcastAddr, 
			sizeof(broadcastAddr), &signal, NULL );
		if ( 0 == nRet )	// Congratulations! You've got a big surprise!
		{
			// Data sent. Continue
			LeaveCriticalSection( &m_QueueLocks[m_nQueueHead] );
		}
		else	// It should be considered normal, while the former one.... Who knows?
		{
			if ( WSA_IO_PENDING != WSAGetLastError() )	// Some unexpected error occurs
			{
				// Don't know how to handle
				// Release it anyway to avoid deadlock
				LeaveCriticalSection( &m_QueueLocks[m_nQueueHead] );
			}
			else	// I love such situation!
			{
				while( WAIT_OBJECT_0 != (nWaitResult = WaitForSingleObject( signal.hEvent, NQDS_TIMEOUT )) )
				{
					if ( m_bTerminateSignal )
					{
						TRACE( "Error waiting in Broadcast thread!\n" );
						break;
					}
				}
				if ( WAIT_OBJECT_0 == nWaitResult )	// IO completes successfully
				{
					// Data sent. Continue with next packet
					LeaveCriticalSection( &m_QueueLocks[m_nQueueHead] );
				}
			}
		}
		EnterCriticalSection( &m_hCriticalSection );
		m_nQueueHead++;
		if ( BROADCAST_QUEUE_SIZE == m_nQueueHead )
			m_nQueueHead = 0;
		LeaveCriticalSection( &m_hCriticalSection );
	}
}

// Sequence of sending a packet:
// 1: PreparePacket()	-----+
// 2: PacketPointer()     <--|
// 3: EndPacket();		  <--+
int CBroadcastThread::PreparePacket()
{
	int retVal;
	EnterCriticalSection( &m_hCriticalSection );
	EnterCriticalSection( &m_QueueLocks[m_nQueueTail] );
	retVal = m_nQueueTail;
	m_nQueueTail++;
	if ( BROADCAST_QUEUE_SIZE == m_nQueueTail )
		m_nQueueTail = 0;
	if ( m_nQueueTail == m_nQueueHead )
	{
		if ( 0 == m_nQueueTail )
			m_nQueueTail = BROADCAST_QUEUE_SIZE-1;
		else
			m_nQueueTail--;
		LeaveCriticalSection( &m_QueueLocks[m_nQueueTail] );
		retVal = -1;
	}
	LeaveCriticalSection( &m_hCriticalSection );
	return retVal;
}

void CBroadcastThread::EndPacket( int nIndex )
{
	if ( -1 != nIndex )
	{
		LeaveCriticalSection( &m_QueueLocks[nIndex] );
		ReleaseSemaphore( this->m_hSemaphore, 1, NULL );
	}
}

void CBroadcastThread::ExitInstance()
{
	closesocket( this->m_hSocket );
	WSACloseEvent( this->m_hIOEvent );	//CloseHandle
}

void CBroadcastThread::SafeTerminateAndCleanUp()
{
	SignalTerminate();
	while ( IsAlive() )
		Sleep(1);
	CloseHandle(m_hThread);
	TRACE( "Broadcast thread terminates!\n" );
	m_bTerminated = TRUE;
}

BOOL CBroadcastThread::IsAlive()
{
	if ( NULL == m_hThread )
		return FALSE;
	DWORD waitResult = WaitForSingleObject( this->m_hThread, 0 );
	if ( WAIT_TIMEOUT == waitResult )
	{
		return TRUE;
	}
	else if ( WAIT_OBJECT_0 == waitResult )
	{
		return FALSE;
	}
	else // Error occures
	{
		return FALSE;
	}
}

static DWORD WINAPI BroadcastThreadProc( LPVOID lpParameter )
{
	CBroadcastThread *thisThread = (CBroadcastThread *)lpParameter;
	thisThread->InitInstance();
	thisThread->ExitInstance();
	return 0;
}
